# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import gzip
import struct

from .constants import *
from .errors import *
from .logging import *
from .types import *
from .utils import *

from ._compat import *


__all__ = ['file_exists', 'file_permissions', 'home_file', 'FileReader', 'ArchiveFileReader',
    'Reader', 'Writer']


def file_exists(path):
    try:
        with open(path):
            return True
    except (OSError, IOError):
        return False

def file_permissions(path):
    return os.stat(path).st_mode & 0o777

def home_file(filename):
    return os.path.join(os.path.expanduser("~"), filename)


class FileReader(object):
    def __init__(self, path, delimiter=None, encoding=None):
        self.delimiter = delimiter
        abspath = os.path.abspath(path)
        if not os.path.exists(os.path.dirname(abspath)):
            raise FileNotFound("Path '{}' does not exist.".format(os.path.dirname(abspath)))
        with open(path, "rb") as f:
            data = f.read(2)
            if data == GZIP_MAGIC:
                _open = gzip.open
            else:
                _open = open
        self.fd = _open(abspath, "r")

    @classmethod
    def check_length(cls, path, length):
        with cls(path) as f:
            for i, line in enumerate(f, 1):
                if i >= length:
                    return True
            return False

    @classmethod
    def check_delimiter(cls, path):
        with cls(path, "rb") as f:
            header = f.readline().strip()
            return infer_delimiter(header)

    def field_names(self):
        header = self.fd.readline().strip()
        if not self.delimiter:
            self.delimiter = infer_delimiter(header)
        return [n.lower().strip() for n in header.split(self.delimiter)]

    @property
    def file_type(self):
        return DELIMITED_TEXT_FILE

    def read(self, size=-1):
        return self.fd.read(size)

    def readline(self):
        return next(self.fd)

    @classmethod
    def read_all(cls, path):
        with cls(path, "rb") as f:
            return f.read()

    @classmethod
    def read_header(cls, path):
        with cls(path) as f:
            return f.read(len(GIRAFFE_MAGIC))

    def __iter__(self):
        while True:
            line = self.readline()
            if line is None:
                break
            yield line

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if not self.fd.closed:
            self.fd.close()


class ArchiveFileReader(FileReader):
    @property
    def file_type(self):
        return GIRAFFE_ARCHIVE_FILE

    def columns(self):
        if self.fd.tell() != 0:
            raise GiraffeError("Cannot read columns, file descriptor must be at 0.")
        self.fd.read(len(GIRAFFE_MAGIC))
        header = self.readline()
        return Columns.deserialize(header)

    def field_names(self):
        return self.columns().names

    def readline(self):
        data = self.fd.read(2)
        if not data:
            return
        row_length = struct.unpack("H", data)[0]
        return self.fd.read(row_length)


class Reader(object):
    def __new__(cls, path, **kwargs):
        data = FileReader.read_header(path)
        if data == GIRAFFE_MAGIC:
            return ArchiveFileReader(path, **kwargs)
        return FileReader(path, **kwargs)


class Writer(object):
    def __init__(self, path=None, mode='', archive=False, use_gzip=False):
        if path is None:
            self.name = "stdout"
            if 'b' in mode:
                self.fd = os.fdopen(sys.stdout.fileno(), 'wb')
            else:
                self.fd = sys.stdout
        else:
            self.name = os.path.basename(path)
            abspath = os.path.abspath(path)
            if not os.path.exists(os.path.dirname(abspath)):
                os.makedirs(os.path.dirname(abspath))
            if archive:
                mode = 'b'
            if use_gzip:
                _open = gzip.open
            else:
                _open = open
            self.fd = _open(abspath, "w{}".format(mode))
            if archive:
                self.write(GIRAFFE_MAGIC)
                self.writen = self.write

    @property
    def is_stdout(self):
        return self.fd.fileno() == 1

    def write(self, buf):
        try:
            self.fd.write(buf)
        except UnicodeEncodeError as error:
            self.fd.write(buf.encode('UTF-8'))

    def writen(self, buf):
        self.write(buf)
        self.write("\n")

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if not self.fd.closed and not self.is_stdout:
            log.debug("Debug[2]", "Closing file descriptor '{}' ...".format(self.fd))
            self.fd.close()
            log.debug("Debug[2]", "File descriptor '{}' closed.".format(self.fd))



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re


__all__ = ['parse_statement', 'prepare_statement', 'remove_curly_quotes']


def parse_statement(s):
    NORMAL = 1
    COMMENT = 2
    QUOTE = 3
    BLOCK_COMMENT = 4

    output = []
    statement = ""
    mode = NORMAL

    for i, c in enumerate(s):
        if mode == COMMENT:
            if c == "\n":
                mode = NORMAL
                statement += " "
        elif mode == BLOCK_COMMENT:
            if c == "/" and i > 0 and s[i-1] == "*":
                mode = NORMAL
        elif mode == QUOTE:
            if c == "'":
                mode = NORMAL
            statement += c
        elif c == "'":
            mode = QUOTE
            statement += c
        elif c == ";":
            output.append(statement.strip())
            statement = ""
        elif c == "-" and i + 1 < len(s) and s[i+1] == "-":
            mode = COMMENT
        elif c == "/" and i + 1 < len(s) and s[i+1] == "*":
            mode = BLOCK_COMMENT
        elif c == "\n" or c == "\r":
            statement += " "
        else:
            statement += c.lower()
    
    statement = statement.strip()
    if statement:
        output.append(statement)
        
    return output

def prepare_statement(statement):
    statement = remove_curly_quotes(statement)
    statements = parse_statement(statement)
    return ";".join(statements)

def remove_curly_quotes(statement):
    return re.sub('(’|‘|“|”)', "'", statement)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import math
import struct

from .constants import *
from .errors import *
from .fmt import *
from .logging import *

from ._compat import *


__all__ = ['Column', 'Columns', 'GiraffeBytes', 'GiraffeDate', 'GiraffeTime', 'GiraffeTimestamp',
    'Result', 'Results', 'Row']


class Column(object):
    """
    An object containing the column information used by Teradata for
    encoding, decoding, and identifying data.

    Implements the :code:`__eq__` magic method for comparison with other instances
    (validating that each field holds the same data), and the :code:`__str__` and
    :code:`__repr__` magic methods for coersion to string values and more pleasant
    visual presentation.

    **Fields:**

    name
        The original name of the column in the table it belongs to
    alias
        The alias of the column in the query
    title
        The title of the column in the query
    type
        Numeric value used by Teradata to indicate column type
    length
        Length in bytes that the field's values occupy
    precision
        Precision of floating-point numeric columns
    scale
        Scale of floating-point numeric columns
    filler
        Indicator that the column belongs to the target table, but
        will not be used in the current operation
    """

    def __init__(self, column, filler=False):
        if isinstance(column, tuple):
            if len(column) in {5, 7}:
                column = {"name": column[0], "type": column[1], "length": column[2], "precision": column[3], "scale": column[4]}
            elif len(column) == 8:
                column = {"name": column[0], "type": column[1], "length": column[2], "precision": column[3], "scale": column[4], "nullable": column[5], "default": column[6], "format": column[7]}
            else:
                raise GiraffeTypeError("Column information is not valid")
        self.original_name = str(ensure_str(column.get("name"))).strip('"').lower()
        self.alias = str(column.get("alias", "")).strip('"').lower()
        self.title = str(column.get("title", "")).strip('"').lower()
        self.type = column.get("type")
        if self.type in cmd_type_map:
            self.tpt_type = self.type
            self.type = cmd_type_map.get(self.type)
        else:
            self.tpt_type = tpt_type_map.get(self.type)
        self.gd_type = gd_type_map.get(self.type, None)
        if self.type is None or self.gd_type is None:
            raise GiraffeTypeError("Cannot determine data type")
        self.length = column.get("length")
        self.precision = column.get("precision")
        self.scale = column.get("scale")
        self._nullable = column.get("nullable", None)
        self._default = column.get("default", None)
        self._format = column.get("format", None)
        self.filler = filler
        self.order = False

    @property
    def name(self):
        if self.title:
            return self.title
        return self.original_name

    @property
    def as_tuple(self):
        return (self.safe_name, self.tpt_type, self.length,
                self.precision, self.scale)

    @property
    def default(self):
        if self._default is None:
            raise GiraffeError("Default information not available")
        return self._default

    @property
    def format(self):
        if self._format is None:
            raise GiraffeError("Default information not available")
        return self._format

    @property
    def nullable(self):
        if self._nullable is None:
            raise GiraffeError("Nullable information not available")
        return self._nullable

    @property
    def safe_name(self):
        return safe_name(self.name)

    def __eq__(self, other):
        return str(self) == str(other)

    def __repr__(self):
        type_name = COL_TYPE_NAMES.get(self.type, "unknown")
        type_name = type_name.replace('_NN', '').replace('_N' , '').lower()
        if self.type in DECIMAL_TYPES :
            type_name = "{}[{}]({},{})".format(type_name, self.length, self.precision, self.scale)
        else:
            type_name = "{}({})".format(type_name, self.length)
        options = []
        if self._default:
            options.append(self._default)
        if self._nullable:
            options.append(self._nullable)
        if self._format:
            options.append(self._format)
        if options:
            type_name = "{} {}".format(type_name, " ".join(options))
        if self.original_name:
            name = self.original_name
        elif self.alias:
            name = self.alias
        elif self.title:
            name = self.title
        else:
            name = "n/a"
        return "Column({} {})".format(safe_name(name), type_name)

    def __str__(self):
        return self.__repr__()


class Columns(object):
    """
    A set of :class:`~giraffez.types.Column` objects, used to convey
    original table schema without losing information.
    """

    def __init__(self, items=[]):
        self.columns = []
        for item in items:
            self.columns.append(Column(item))
        self._filtered_columns = []
        self._column_map = {}
        for index, column in enumerate(self.columns):
            self._column_map[column.safe_name] = index
            if column.alias:
                self._column_map[safe_name(column.alias)] = index
            if column.title:
                self._column_map[safe_name(column.title)] = index

    def append(self, item):
        if isinstance(item, Column):
            self.columns.append(item)
        else:
            self.columns.append(Column(item))

    def get(self, column_name):
        """
        Retrieve a column from the list with name value :code:`column_name`

        :param str column_name: The name of the column to get
        :return: :class:`~giraffez.types.Column` with the specified name, or :code:`None` if it does not exist.
        """
        column_name = column_name.lower()
        for c in self.columns:
            if c.name == column_name:
                return c
        return None

    @property
    def header_length(self):
        return int(math.ceil(len(self)/8.0))

    @property
    def names(self):
        """
        :return: The names of the contained :class:`~giraffez.types.Column` objects
        :rtype: list
        """
        return [c.name for c in self]

    @property
    def safe_names(self):
        return [c.safe_name for c in self]

    def set_filter(self, names=None):
        """
        Set the names of columns to be used when iterating through the list,
        retrieving names, etc.

        :param list names: A list of names to be used, or :code:`None` for all
        """
        _names = []
        if names:
            for name in names:
                _safe_name = safe_name(name)
                if _safe_name not in self._column_map:
                    raise GiraffeTypeError("Column '{}' does not exist".format(name))
                if _safe_name in _names:
                    continue
                _names.append(_safe_name)
        self._filtered_columns = _names

    def tuples(self):
        return [c.as_tuple for c in self]

    def __eq__(self, other):
        return all([a == b for a, b in zip(self, other)])

    def __getitem__(self, index):
        if self._filtered_columns:
            return self.get(self._filtered_columns[index])
        return self.columns[index]

    def __iter__(self):
        if self._filtered_columns:
            for c in self._filtered_columns:
                column = self.get(c)
                if column is None:
                    raise GiraffeTypeError("Column '{}' does not exist".format(c))
                yield column
        else:
            for c in self.columns:
                yield c

    def __len__(self):
        return len(self.names)

    def __str__(self):
        return ",".join([c.name for c in self.columns])

    def __repr__(self):
        return "\n".join([repr(c) for c in self.columns])

    def serialize(self):
        """
        Serializes the columns into the giraffez archive header
        binary format::

            0      1      2      
            +------+------+------+------+------+------+------+------+
            | Header      | Header Data                             |
            | Length      |                                         |
            +------+------+------+------+------+------+------+------+

                       giraffez Archive Header Format

                                   Fig. 1

            Header Length: 2 bytes
                Full length of archive header 

            Header Data: variable
                Binary data representing N column(s) using the format
                specified in Fig. 2


            0      1      2      3      4      5      6      7      8
            +------+------+------+------+------+------+------+------+
            | Type        | Length      | Precision   | Scale       | 
            +------+------+------+------+------+------+------+------+
            | Name Length | Name                                    |
            +------+------+------+------+------+------+------+------+

                           Binary Column Format

                                   Fig. 2

            Type: 2 bytes
                Numerical representation of column type defined by
                Teradata's CLIv2

            Length: 2 bytes
                Column length

            Scale: 2 bytes
                Column scale

            Precision: 2 bytes
                Column precision

            Name Length: 2 bytes
                Length used for reading variable column name

            Name: variable
                Name of column

        :return: Packed binary data, representing the serialized :class:`~giraffez.types.Columns`
        :rtype: str
        """
        data = b""
        for column in self:
            row = struct.pack("5H", column.type, column.length, column.precision, column.scale,
                len(column.name))
            row += ensure_bytes(column.name)
            data += row
        return struct.pack("H", len(data)) + data

    @classmethod
    def deserialize(cls, data):
        """
        Deserializes giraffez Archive header. See
        :meth:`~giraffez.types.Columns.serialize` for more information.

        :param str data: data in giraffez Archive format, to be deserialized
        :return: :class:`~giraffez.types.Columns` object decoded from data
        """
        column_list = cls()
        while data:
            tup, data = data[:10], data[10:]
            column_type, length, prec, scale, title_len = struct.unpack("5H", tup)
            title, data = data[:title_len], data[title_len:]
            try:
                column_list.append((title, column_type, length, prec, scale))
            except GiraffeTypeError as error:
                raise GiraffeEncodeError(error)
        return column_list
        

class GiraffeBytes(object):
    def __init__(self, columns):
        self.raw_array = bytearray([0] * columns.header_length)

    @classmethod
    def read(cls, f):
        for b in iterbytes(f):
            for i in reversed(xrange(8)):
                yield (b >> i) & 1

    def set_bit(self, offset):
        byte, bit = divmod(offset, 8)
        self.raw_array[byte] |= 128 >> bit

    def __add__(self, other):
        return b2s(self) + other

    def __bytes__(self):
        return bytes(self.raw_array)

    def __str__(self):
        return str(self.raw_array)


class GiraffeDate(datetime.datetime):
    """
    Ensures that datetime objects can be proper coerced into a
    string value of a given format.

    There is a Python bug (yes, it is a bug, not a feature) where
    datetime objects representing any dates before 1900 raises an
    error when calling the strftime method. This will ensure that the
    date is coerced safely with str() into something Teradata can use.
    """

    def __init__(self, *args, **kwargs):
        super(GiraffeDate, self).__init__()

    @classmethod
    def from_datetime(cls, dt):
        if isinstance(dt, datetime.datetime):
            return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
        else:
            return cls(dt.year, dt.month, dt.day, 0, 0, 0)

    @classmethod
    def from_integer(cls, d):
        try:
            if not d:
                return None
            return cls.strptime(str(d + 19000000), "%Y%m%d")
        except ValueError as error:
            log.debug(error)
            return None

    @classmethod
    def from_string(cls, s):
        for fmt in DATE_FORMATS:
            try:
                return cls.strptime(str(s), fmt)
            except ValueError as error:
                pass
        return None

    def to_integer(self):
        try:
            date_value = "{0:04}{1:02}{2:02}".format(self.year, self.month, self.day)
            return int(date_value) - 19000000
        except ValueError as error:
            return None

    def to_json(self):
        return unicode(self.to_string())

    def to_string(self):
        return "{0:04}-{1:02}-{2:02}".format(self.year, self.month, self.day)

    def __bytes__(self):
        return bytes(self.to_string(), "UTF-8")

    def __int__(self):
        return self.to_integer()

    def __str__(self):
        return self.to_string()

    def __unicode__(self):
        return unicode(self.to_string())


class GiraffeTime(datetime.time):
    """
    Represents Teradata date/time data types such as TIME(0). Currently,
    does not keep microsecond for other types like TIME(n).
    """
    def __new__(cls, t):
        return datetime.time.__new__(cls, t.hour, t.minute, t.second, t.microsecond)

    @classmethod
    def from_string(cls, s):
        fmt = TIME_FORMATS.get(len(s.strip()))
        if fmt is not None:
            try:
                ts = datetime.datetime.strptime(str(s), fmt)
                return cls(datetime.time(ts.hour, ts.minute, ts.second, ts.microsecond))
            except ValueError as error:
                log.debug("GiraffeTime: ", error)
        return None


class GiraffeTimestamp(GiraffeDate):
    """
    Represents Teradata date/time data types such as TIMESTAMP(n).
    """

    def __init__(self, *args, **kwargs):
        super(GiraffeTimestamp, self).__init__(*args, **kwargs)
        self._original_length = None

    @classmethod
    def from_string(cls, s):
        format = TIME_FORMATS.get(len(s.strip()))
        if format is not None:
            try:
                ts = cls.strptime(str(s), format)
                ts._original_length = len(s)
                return ts
            except ValueError as error:
                log.debug("GiraffeTimestamp: ", error)
        return None

    def to_string(self, length=None):
        if length is None:
            length = self._original_length
        if length is not None and length > 20:
            value = "{0:04}-{1:02}-{2:02} {3:02}:{4:02}:{5:02}.{6:06}".format(self.year, self.month, self.day,
                    self.hour, self.minute, self.second, self.microsecond)
            value = value[:length]
            return value
        return "{0:04}-{1:02}-{2:02} {3:02}:{4:02}:{5:02}".format(self.year, self.month, self.day,
                    self.hour, self.minute, self.second)


class Result(object):
    """
    Contains the result of a Teradata query

    Defines the :code:`__len__` magic method to get the number of rows returned:

    .. code-block:: python

       result = cli.execute("select * from tablename;")
       num_rows = len(result)

    Defines the :code:`__iter__` magic method to iterate through the returned rows:

    .. code-block:: python

       for row in result:
           print row

    Defines the :code:`__getattr__` magic method to return a list of values for a given column:

    .. code-block:: python

       print results.first_name # ['alice', 'bob', 'charlie', ...]

    Defines the :code:`__getitem__` magic method to return a list of values for a given column, or numeric index:

    .. code-block:: python

       print results.columns.names # first_name, last_name, id
       print results['first_name'] # ['alice', 'bob', 'charlie', ...]
       print results['id'] # ['abc123', 'bcd234', 'cde345', ...]
       print results[2] # ['abc123', 'bcd234', 'cde345', ...]
    """

    def __init__(self, result):
        self.result = result

    @property
    def columns(self):
        """
        The column data for the corresponding rows

        :rtype: :class:`~giraffez.types.Columns` 
        """
        return self.result["columns"]

    def first(self):
        """
        Return the first row (or :code:`None`) if no rows in result

        :return: The first row in the results
        :rtype: :class:`~giraffez.types.Row`
        """
        try:
            return self.result["rows"][0]
        except IndexError:
            return None

    @property
    def rows(self):
        """
        The returned row data

        :return: a list of rows
            (ordered corresponding to :meth:`~giraffez.types.Result.columns`)
        :rtype: list of :class:`~giraffez.types.Row`
        """
        return self.result["rows"]

    def items(self):
        """
        Generator method for iterating over the :code:`dict` representations of
        each row in the result

        :return: a generator yielding each row with 
            :meth:`~giraffez.types.Row.to_dict` called on each
        :rtype: generator of :code:`dict`
        """
        for row in self.rows:
            yield row.to_dict()

    def to_json(self):
        """
        .. deprecated:: 1.0.5
            Use :meth:`~giraffez.types.Result.items` instead.
        """
        return self.items()

    def __getattr__(self, name):
        try:
            index = self.columns._column_map[name]
            return [row.__getitem__(index) for row in self.rows]
        except KeyError:
            raise AttributeError("'{}' object has no attribute '{}'".format(
                self.__class__.__name__, name))

    def __getitem__(self, key):
        if isinstance(key, basestring):
            try:
                index = self.columns._column_map[key]
                return [row.__getitem__(index) for row in self.rows]
            except KeyError:
                raise AttributeError("'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, key))
        else:
            return self.rows.__getitem__(key)

    def __iter__(self):
        for row in self.result["rows"]:
            yield row

    def __len__(self):
        return len(self.result["rows"])

    def __repr__(self):
        return "\n".join([repr(x) for x in self.rows])

    def __str__(self):
        return self.__repr__()


class Results(object):
    """
    A set of :class:`~giraffez.types.Result` objects returned by queries

    Defines the :code:`__iter__` magic method for convenience:

    .. code-block:: python

       results = cli.execute_many('''select * from table_a;
           select * from table_b;''')
       for result in results:
           print result.rows

    Defines the :code:`__getitem__` magic method to allow numeric indexing:

    .. code-block:: python

       results = cli.execute_many('''select * from table_a;
           select * from table_b;''')
       print results[1] # results from 'select * from table_b'
    """

    def __init__(self, results=[]):
        self.results = []
        for result in results:
            self.append(result)

    def append(self, item):
        if isinstance(item, Result):
            self.results.append(item)
        elif isinstance(item, dict):
            self.results.append(Result(item))
        else:
            raise GiraffeError(("Object of type '{}' is not a valid argument for "
                "`Results.append`.").format(type(item)))

    def one(self):
        """
        Return the first result, or none if there are no result sets

        :return: The first result in the set
        :rtype: :class:`~giraffez.types.Result`
        """
        try:
            return self.results[0]
        except IndexError:
            return None

    def __getitem__(self, key):
        return self.results.__getitem__(key)

    def __iter__(self):
        for result in self.results:
            yield result

    def __len__(self):
        return len(self.results)

    def __repr__(self):
        return "\n".join([str(x) for x in self.results])

    def __str__(self):
        return self.__repr__()


class Row(object):
    """
    A wrapper for a single row within a :class:`~giraffez.types.Result` object.

    Defines the :code:`__iter__` magic method for convenience:

    .. code-block:: python

       row = results.first()
       for item in row:
           print item

    Defines the :code:`__getattr__` magic method to return a particular field:

    .. code-block:: python

       row = results.first()
       print row.first_name # 'alice'

       for row in results:
           print row.first_name # 'alice', 'bob', 'charlie', ...

    Defines the :code:`__getitem__` magic method to return a particular field by name or numeric index:

    .. code-block:: python

       row = results.first()
       print row['first_name'] # 'alice'
       print row['id'] # 'abc123'
       print row[2] # 'abc123'
    """

    def __init__(self, columns, row):
        self.columns = columns
        self.column_map = columns._column_map
        self.row = row

    def to_dict(self):
        """
        Represents the contents of the row as a :code:`dict` with the column
        names as keys, and the row's fields as values. Used by 
        :meth:`~giraffez.types.Result.items` and useful for passing row data to
        functions which expect dictionaries.

        :rtype: dict
        """
        return {k.name: v for k, v in zip(self.columns, self)}

    def to_json(self):
        """
        .. deprecated:: 1.0.5
            Use :meth:`~giraffez.types.Row.to_dict` instead.
        """
        return self.to_dict()

    def __getattr__(self, name):
        try:
            index = self.column_map[name]
            return self.__getitem__(index)
        except KeyError:
            raise AttributeError("'{}' object has no attribute '{}'".format(
                self.__class__.__name__, name))

    def __getitem__(self, key):
        if isinstance(key, basestring):
            try:
                return self.row[self.column_map[key]]
            except KeyError:
                raise AttributeError("Row has no column '{}'".format(key))
        else:
            return self.row.__getitem__(key)

    def __iter__(self):
        for item in self.row:
            yield item

    def __len__(self):
        return len(self.row)

    def __repr__(self):
        return str(self.to_dict())

    def __str__(self):
        return self.__repr__()



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from collections import defaultdict

from .cmd import *
from .connection import *
from .constants import *
from .encoders import *
from .errors import *
from .io import *
from .logging import *
from .sql import *
from .types import *
from .utils import *


__all__ = ['TeradataLoad']


class TeradataLoad(TeradataCmd):
    """
    The class for inserting into Teradata tables using CLIv2. 

    Exposed under the alias :class:`giraffez.Load`.
    
    This class should be used for inserting into Teradata tables in all but
    very large (> ~100k rows) cases.

    See :class:`~giraffez.cmd.TeradataCmd` for constructor arguments.

    Meant to be used, where possible, with python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete:

    .. code-block:: python

       with giraffez.Load() as load:
           load.from_file('myfile.txt', 'database.my_table')
           # continue executing statements and processing results

    Use in this manner guarantees proper exit-handling and disconnection 
    when operation is completed (or interrupted).
    """

    def from_file(self, table_name, input_file_name, delimiter=None, null=DEFAULT_NULL,
            date_conversion=False):
        """
        Load a text file into the specified :code:`table_name`

        For most insertions, this will be faster and produce less strain on
        Teradata than using :class:`~giraffez.load.TeradataMLoad` (:class:`giraffez.MLoad`).

        Requires that the input file be a properly delimited text file, with a
        header that corresponds to the target fields for insertion. Valid delimiters
        include '|', ',', and '\\t' (tab).

        :param str table_name: The name of the destination table
        :param str input_file_name: The name of the file to read rows from
        :param str delimiter: The delimiter used by the input file (or :code:`None` 
            to infer it from the header).
        :param str null: The string used to indicated nulled values in the
            file (defaults to :code:`'NULL'`).
        :param bool date_conversion: If :code:`True`, attempts to coerce date fields
            into a standard format (defaults to :code:`False`).
        :return: A dictionary containing counts of applied rows and errors
        :rtype: dict
        """
        with Reader(input_file_name, delimiter=delimiter) as f:
            fields = f.field_names()
            preprocessor = null_handler(null)
            rows = (preprocessor(l.strip().split(f.delimiter)) for l in f)
            return self.insert(table_name, rows, fields=fields, date_conversion=date_conversion)

    def insert(self, table_name, rows, fields=None, date_conversion=True):
        """
        Insert Python :code:`list` rows into the specified :code:`table_name`

        :param str table_name: The name of the destination table
        :param list rows: A list of rows. Each row must be a :code:`list` of 
            field values.
        :param list fields: The names of the target fields, in the order that 
            the data will be presented (defaults to :code:`None` for all columns in the table).
        :param bool date_conversion: If :code:`True`, attempts to coerce date fields
            into a standard format (defaults to :code:`True`).
        :return: A dictionary containing counts of applied rows and errors
        :rtype: dict
        :raises `giraffez.errors.GiraffeEncodeError`: if the number of values in a row does not match 
            the length of :code:`fields`
        :raises `giraffez.errors.GiraffeError`: if :code:`panic` is set and the insert statement 
            caused an error.
        """
        columns = self.get_columns(table_name)
        if fields is None:
            fields = columns.safe_names
        columns.set_filter(fields)
        check_input(columns, fields)
        stats = defaultdict(int)
        processor = pipeline([
            python_to_sql(table_name, columns, date_conversion)
        ])
        def _fetch():
            stats['count'] = 0
            current_block = ""
            for row in rows:
                try:
                    stmt = processor(row)
                except GiraffeError as error:
                    if self.panic:
                        raise error
                    log.info("Load", error)
                    stats['errors'] += 1
                    continue
                if len(current_block + stmt) > CLI_BLOCK_SIZE:
                    yield current_block
                    current_block = ""
                current_block += stmt
                stats['count'] += 1
            if current_block:
                yield current_block
        for block in _fetch():
            self.execute_many(block, sanitize=True, parallel=True)
        return stats



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class GiraffeError(Exception):
    """
    Baseclass for all giraffez errors.
    """

class GeneralError(GiraffeError):
    """
    General giraffez error.
    """

class GiraffeNotFound(GiraffeError):
    """
    Raised when giraffez C module not found.
    """

class GiraffeTypeError(GiraffeError):
    """
    Baseclass for all giraffez type errors.
    """

class GiraffeEncodeError(GiraffeError):
    """
    Raised when unable to encode the provided object.
    """

class ConfigurationError(GiraffeError): 
    """
    For use with configuration file handling. 
    """

class FileNotFound(GiraffeError): 
    """
    Raised when file does not exist.
    """

class ConfigNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class KeyNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class ConfigReadOnly(ConfigurationError):
    """
    Raised when a write is attempted on a configuration file was opened
    in read mode.
    """

class ConnectionLock(ConfigurationError):
    """
    Raised when connection is locked by invalid attempts and the
    'protect' feature is being used.
    """

    def __init__(self, dsn):
        super(ConnectionLock, self).__init__(("Connection {0} is currently locked. please update "
            "credentials and run:\n\tgiraffez config --unlock {0}").format(dsn))

class TeradataError(Exception): 
    """
    Baseclass for all Teradata errors. This exception represents any
    error that originates from Teradata CLIv2/TPTAPI.

    This will attempt to parse error codes from the message in the case
    that the error is a Teradata CLIv2 error.
    """

    def __init__(self, message, code=None):
        super(TeradataError, self).__init__(message)

        if isinstance(message, Exception):
            message = getattr(message, 'message', '')

        if code is None:
            if isinstance(message, (list, tuple)):
                code, message = message[0], message[1]
            else:
                if ":" in message:
                    parts = [x.strip() for x in message.split(":", 1)]
                    try:
                        code = int(parts[0])
                        message = parts[1]
                    except ValueError as error:
                        code = None
                        message = "Unable to parse CLI error code/message"
                else:
                    code = None
                    message = "Unable to parse CLI error code/message"
            if code == ObjectDoesNotExist.code:
                raise ObjectDoesNotExist(message, code)
            elif code == ObjectNotTable.code:
                raise ObjectNotTable(message, code)
            elif code == ObjectNotView.code:
                raise ObjectNotView(message, code)
            elif code == TransactionAborted.code:
                raise TransactionAborted(message, code)
            elif code == CannotReleaseMultiLoad.code:
                raise CannotReleaseMultiLoad(message, code)
            elif code == MultiLoadTableExists.code:
                raise MultiLoadTableExists(message, code)
            elif code == MultiLoadWorkTableNotFound.code:
                raise MultiLoadWorkTableNotFound(message, code)
            elif code == InvalidCredentialsError.code:
                raise InvalidCredentialsError(message, code)

        self.message = message
        self.code = code

    def __repr__(self):
        return "{}: {}".format(self.code, self.message)

class ObjectDoesNotExist(TeradataError):
    """
    Teradata object does not exist.
    """
    code = 3807

class ObjectNotTable(TeradataError):
    """
    Teradata object not a table.
    """
    code = 3853

class ObjectNotView(TeradataError):
    """
    Teradata object not a view.
    """
    code = 3854

class TransactionAborted(TeradataError):
    """
    Teradata transaction aborted.
    """
    code = 2631

class MultiLoadError(TeradataError):
    """
    General MultiLoad error class.
    """

class CannotReleaseMultiLoad(MultiLoadError):
    """
    Raised when MultiLoad cannot be released because it is in the
    apply phase.
    """
    code = 2572

class MultiLoadTableExists(MultiLoadError):
    """
    Raised when MultiLoad worktables exist.
    """
    code = 2574

class MultiLoadLocked(MultiLoadError):
    """
    Raised when MultiLoad appears to be locked. Used to determine when
    MultiLoad should be released.
    """

class MultiLoadWorkTableNotFound(MultiLoadLocked):
    """
    Raised when MultiLoad worktable is missing during restart.
    """
    code = 2583

class InvalidCredentialsError(TeradataError):
    """
    Raised when connection credentials are incorrect.
    """
    code = 8017



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

GIRAFFE_NOT_FOUND = False
try:
    from . import _encoder
except ImportError:
    GIRAFFE_NOT_FOUND = True

TDCLI_NOT_FOUND = False
try:
    from . import _cli
except ImportError:
    TDCLI_NOT_FOUND = True

from .config import *
from .connection import *
from .constants import *
from .encoders import *
from .errors import *
from .fmt import *
from .io import *
from .logging import *
from .parser import *
from .sql import *
from .types import *
from .utils import *


__all__ = ['TeradataCmd']


class TeradataCmd(Connection):
    """
    The class for connecting to Teradata and executing commands and 
    queries using CLIv2. 

    Exposed under the alias :class:`giraffez.Cmd`.
    
    For large-output queries, :class:`giraffez.Export` should be used.

    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE` :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE` and :code:`giraffez.DEBUG`
    :param bool panic: If :code:`True` stop executing commands and return after the first error.
    :param str config: Specify an alternate configuration file to be read from, when previous paramaters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True` 
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>` changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :param string mload_session: Identifies mload sessions to suppress duplicate log output.
        Used internally only.
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.errors.TeradataError`: if the connection cannot be established

    Meant to be used, where possible, with python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete:

    .. code-block:: python

       with giraffez.Cmd() as cmd:
           results = cmd.execute('select * from dbc.dbcinfo')
           # continue executing statements and processing results

    Use in this manner guarantees proper exit-handling and disconnection 
    when operation is completed.
    """

    def __init__(self, host=None, username=None, password=None, log_level=INFO, panic=False,
            config=None, key_file=None, dsn=None, protect=False, mload_session=False):
        if TDCLI_NOT_FOUND:
            raise GiraffeNotFound("giraffez module was not compiled with package")
        super(TeradataCmd, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect, mload_session)

        self.panic = panic

    def _connect(self, host, username, password):
        self.cmd = _cli.Cmd()
        try:
            self.cmd.connect(host, username, password)
        except _cli.error as error:
            raise TeradataError(error)

    def _execute(self, command):
        try:
            results, error = self.cmd.execute(command)
            if error is not None:
                raise TeradataError(error)
            log.debug("Debug[2]", repr(results))
        except _cli.error as error:
            raise TeradataError(error)
        for result in results:
            stmt_info = result.get('stmtinfo', None)
            if not stmt_info:
                raise GiraffeError("Statement Info header not received")
            columns = _encoder.Encoder.unpack_stmt_info(stmt_info)
            result["columns"] = Columns(columns)
            for column in result["columns"]:
                log.verbose("Debug[1]", repr(column))
            encoder = _encoder.Encoder(result["columns"])
            processor = pipeline([
                date_handler(result["columns"]),
                char_handler(result["columns"]),
                lambda s: Row(result["columns"], s)
            ])
            result["rows"] = [processor(row) for row in encoder.unpack_rows(result["rows"])]
        return results

    def close(self):
        if getattr(self, 'cmd', None):
            self.cmd.close()

    def collect_stats(self, table_name, predicate=""):
        """
        Collect statistics on the table :code:`table_name` using the given :code:`predicate` if present

        :param str table_name: The name of the table to collect statistics on
        :param str predicate: The predicate of the collect statistics statement, which follows the table name
        :return: The result of the collect statistics statement
        :rtype: None or :class:`~giraffez.types.Result`
        :raises `giraffez.errors.TeradataError`: if statistic cannot be collected for :code:`table_name`
        :raises `giraffez.errors.ObjectDoesNotExist`: if the given :code:`table_name` does not exist
        """
        self.execute_one("collect statistics on {} {}".format(table_name, predicate))

    def count_table(self, table_name):
        """
        Return the row count of :code:`table_name`

        :param str table_name: The name of the table to be counted
        :return: The number of rows in :code:`table_name`
        :rtype: int
        """
        return self.execute_one("select count(*) from {}".format(table_name)).first()[0]

    def drop_table(self, table_name):
        """
        Execute the statement :code:`drop table table_name`

        :param str table_name: The name of the table to be dropped
        :return: :code:`None` if :code:`table_name` does not exist, otherwise the result of
            the statement
        :rtype: None or :class:`~giraffez.types.Result`
        """
        try:
            result = self.execute_one("drop table {}".format(table_name))
        except ObjectDoesNotExist as error:
            return None
        return result

    def execute(self, command, sanitize=True):
        """
        Execute commands using CLIv2.

        :param str command: The SQL command to be executed
        :param bool sanitize: Whether or not to call :func:`~giraffez.sql.prepare_statement` on the command
        :return: the results of each statement in the command
        :rtype: :class:`~giraffez.types.Results`
        :raises `giraffez.errors.TeradataError`: if the query is invalid
        :raises `giraffez.errors.GiraffeError`: if the return data could not be decoded
        """
        self.options("panic", self.panic)
        if log.level >= VERBOSE:
            self.options("query", command, 2)
        else:
            self.options("query", truncate(command), 2)
        log.info("Command", "Executing ...")
        log.info(self.options)
        if sanitize:
            command = prepare_statement(command) # accounts for comments and newlines
            log.debug("Debug[2]", "Command (sanitized): {!r}".format(command))
        return Results(self._execute(command))

    def execute_one(self, command, sanitize=True):
        """
        Execute a single command using CLIv2.

        :param str command: The SQL command to be executed
        :param bool sanitize: Whether or not to call :func:`~giraffez.utils.clean_query` on the command
        :return: the result set of the first statement in the command (all are executed)
        :rtype: :class:`~giraffez.types.Result`
        :raises `giraffez.errors.TeradataError`: if the query is invalid
        :raises `giraffez.errors.GiraffeError`: if the return data could not be decoded
        """
        return self.execute(command, sanitize).one()

    def execute_many(self, command, sanitize=True, parallel=False):
        """
        Execute multiple commands using CLIv2, splitting on statements and executing them individually. With :code:`parallel` set, the behavior becomes that of :meth:`~giraffez.Cmd.execute`.

        :param str command: The SQL command(s) to be executed
        :param bool sanitize: Whether or not to call :func:`~giraffez.utils.clean_query` on the statements
        :param bool parallel: If :code:`True` call :meth:`~giraffez.Cmd.execute` instead
        :return: the results of all executed commands
        :rtype: :class:`~giraffez.types.Results`
        """
        if parallel:
            return self.execute(command, sanitize)
        statements = parse_statement(command)
        results = []
        for statement in statements:
            results.append(self.execute_one(statement, sanitize))
        return Results(results)

    def exists(self, object_name):
        """
        Check that object (table or view) :code:`object_name` exists, by executing a :code:`show table object_name` query, 
        followed by a :code:`show view object_name` query if :code:`object_name` is not a table.

        :param str object_name: The name of the object to check for existence.
        :return: :code:`True` if the object exists, :code:`False` otherwise.
        :rtype: bool
        """
        try:
            self.execute("show table {}".format(object_name))
            return True
        except ObjectNotTable as error:
            self.execute("show view {}".format(object_name))
            return True
        except (ObjectNotView, ObjectDoesNotExist) as error:
            return False

    def get_columns(self, table_name):
        """
        Return the column information for :code:`table_name` by executing a :code:`select top 1 * from table_name` query.

        :param str table_name: The fully-qualified name of the table to retrieve schema for
        :return: the columns of the table
        :rtype: :class:`~giraffez.types.Columns`
        """
        return self.execute_one("select top 1 * from {}".format(table_name)).columns

    def release_mload(self, table_name):
        """
        Release an MLoad lock on :code:`table_name` by executing the statement :code:`release mload table_name`
        followed by the statement :code:`release mload table_name in apply` if the lock cannot be released

        :param str table_name: The name of the table to be dropped
        :return: :code:`None` if :code:`table_name` does not exist, otherwise the result of 
            the statement
        :rtype: None or :class:`~giraffez.types.Result`
        """
        try:
            return self.execute_one("release mload {}".format(table_name))
        except CannotReleaseMultiLoad as error:
            return self.execute_one("release mload {} in apply".format(table_name))
